% Possible forms of axioms in OWL
% This defines the set of possible arguments to axiom/1
% This is essentially a translation of the definition of possible
% axioms (OWL language semantics and abstract syntax paper) into Prolog

:- dynamic owl_axiom/1.

owl_axiom(class(ID,Modality,DList)) :-
   class_id(ID), modality(Modality), description_list(DList).
owl_axiom(enumerated_class(ID,IList)) :-
   class_id(ID), individual_idlist(IList).
owl_axiom(disjoint_classes([D1,D2|DList])) :- description_list([D1,D2|DList]).
owl_axiom(equivalent_classes([D1,D2|DList])) :- description_list([D1,D2|DList]).
   % OWL says can have just one!
owl_axiom(subclass_of(D1,D2)) :- description(D1), description(D2).
owl_axiom(datatype(ID)) :- datatype_id(ID).
owl_axiom(datatype_property(DPID,DPL)) :-
   datavalued_property_id(DPID), data_prop_list(DPL).
owl_axiom(object_property(DPID,DPL)) :-
   individualvalued_property_id(DPID), indiv_prop_list(DPL).
owl_axiom(equivalent_properties([D1,D2|DPIDs])) :-
   data_prop_id_list([D1,D2|DPIDs]).   
owl_axiom(subproperty_of(DPID1,DPID2)) :-
   datavalued_property_id(DPID1), datavalued_property_id(DPID2).
owl_axiom(equivalent_properties([D1,D2|DPIDs])) :-
   indiv_prop_id_list([D1,D2|DPIDs]).   
owl_axiom(subproperty_of(DPID1,DPID2)) :-
   individualvalued_property_id(DPID1), individualvalued_property_id(DPID2).
owl_axiom(instance_of(I,C)) :-
   class_id(C), individual_id(I).
owl_axiom(related(I,P,I1)) :-
   individual_id(I), individual_id(I1), individualvalued_property_id(P).

   :- dynamic modality/1.

   modality(complete).
   modality(partial).

:- dynamic description/1.

description('TOP').
description('BOTTOM').
description(ID) :- class_id(ID).
description(R) :- restriction(R).
description(union_of([D1,D2|DList])) :- description_list([D1,D2|DList]).
   % OWL says you can have any number!
description(intersection_of([D1,D2|DList])) :- description_list([D1,D2|DList]).
   % OWL says you can have any number!
description(complement_of(D)) :- description(D).
description(one_of([I|IDList])) :- individual_idlist([I|IDList]).
   % OWL says you can have zero!

:- dynamic restriction/1.

restriction(restriction(DPID,[D|DRCList])) :-
   datavalued_property_id(DPID), drclist([D|DRCList]).
restriction(restriction(DPID,[D|DRCList])) :-
   individualvalued_property_id(DPID), irclist([D|DRCList]).

:- dynamic data_restriction_component/1.

data_restriction_component(all_values_from(R)) :- data_range(R).
data_restriction_component(some_values_from(R)) :- data_range(R).
data_restriction_component(value(R)) :- data_literal(R).
data_restriction_component(C) :- cardinality(C).

:- dynamic individual_restriction_component/1.

individual_restriction_component(all_values_from(D)) :- description(D).
individual_restriction_component(some_values_from(D)) :- description(D).
individual_restriction_component(value(R)) :- individual_id(R).
individual_restriction_component(C) :- cardinality(C).

:- dynamic cardinality/1.

cardinality(min_cardinality(I)) :- integer(I), I>=0.
cardinality(max_cardinality(I)) :- integer(I), I>=0.
cardinality(cardinality(I)) :- integer(I), I>=0.

:- dynamic data_range/1.

data_range(D) :- datatype_id(D).
data_range(one_of(DLList)) :- data_literal_list(DLList).

:- dynamic data_prop_elt/1.

data_prop_elt(super(DPID)) :- datavalued_property_id(DPID).
data_prop_elt(functional).
data_prop_elt(domain(D)) :- description(D).
data_prop_elt(range(R)) :- data_range(R).

:- dynamic indiv_prop_elt/1.

indiv_prop_elt(super(DPID)) :- individualvalued_property_id(DPID).
indiv_prop_elt(inverse_of(DPID)) :- individualvalued_property_id(DPID).
indiv_prop_elt(symmetric).
indiv_prop_elt(functional).
indiv_prop_elt(inverse_functional).
indiv_prop_elt(transitive).
indiv_prop_elt(domain(D)) :- description(D).
indiv_prop_elt(range(D)) :- description(D).

% Lists

   :- dynamic description_list/1, drclist/1, irclist/1.
   :- dynamic individual_idlist/1, data_literal_list/1.
   :- dynamic data_prop_list/1, indiv_prop_list/1, indiv_prop_id_list/1.
   :- dynamic data_prop_id_list/1.

   description_list([]).
   description_list([D|Ds]) :- description(D), description_list(Ds).

   drclist([]).
   drclist([DRC|DRCs]) :- data_restriction_component(DRC), drclist(DRCs).

   irclist([]).
   irclist([DRC|DRCs]) :- individual_restriction_component(DRC), irclist(DRCs).

   individual_idlist([]).
   individual_idlist([ID|IDs]) :- individual_id(ID), individual_idlist(IDs).

    data_literal_list([]).
    data_literal_list([DL|DLs]) :- data_literal(DL), data_literal_list(DLs).

   data_prop_list([]).
   data_prop_list([P|Ps]) :- data_prop_elt(P), data_prop_list(Ps).

   indiv_prop_list([]).
   indiv_prop_list([P|Ps]) :- indiv_prop_elt(P), indiv_prop_list(Ps).

   data_prop_id_list([]).
   data_prop_id_list([P|Ps]) :- datavalued_property_id(P), data_prop_id_list(Ps).

   indiv_prop_id_list([]).
   indiv_prop_id_list([P|Ps]) :- individualvalued_property_id(P), 
      indiv_prop_id_list(Ps).



% IDs and literals

:- dynamic class_id/1, datatype_id/1, data_literal/1, datavalued_property_id/1.
:- dynamic individualvalued_property_id/1, individual_id/1.

class_id(X) :- atom(X).
datatype_id(X) :- atom(X).
data_literal(X) :- atomic(X).
datavalued_property_id(ID) :- atom(ID).
individualvalued_property_id(ID) :- atom(ID).
individual_id(ID) :- atom(ID).

% Utilities for manipulating OWL formulae

   add_union(union_of(L1),union_of(L2),union_of(L3)) :- !, append(L1,L2,L3).
   add_union(union_of(L),X,union_of([X|L])) :- !.
   add_union(X,union_of(L),union_of([X|L])) :- !.
   add_union(X,Y,union_of([X,Y])).

   intersect([X],X) :- !.
   intersect(L,intersection_of(L)).

   conjs_of(intersection_of(L),L) :- !.
   conjs_of(X,[X]).


% write a formula

wax(X) :- atom(X), !, write(X).
wax(intersection_of(List)) :- winter(List), !.
wax(complement_of(X)) :- write('~'), wax(X), !.
wax(union_of(X)) :- write('('), wunion(X), write(')'), !.
wax(restriction(P,[some_values_from(C)])) :- 
   write('(E '), write(P), write('. '), wax(C), write(')'), !.
wax(restriction(P,[all_values_from(C)])) :- 
   write('(A '), write(P), write('. '), wax(C), write(')'), !. 
wax(restriction(P,[cardinality(C)])) :- 
   write('(= '), write(C), write(' '), write(P), write(')'), !.
wax(restriction(P,[min_cardinality(C)])) :- 
   write('(]= '), write(C), write(' '), write(P), write(')'), !.
wax(restriction(P,[max_cardinality(C)])) :- 
   write('(=[ '), write(C), write(' '), write(P), write(')'), !.
wax(restriction(P,[min_cardinality(C1),max_cardinality(C2)])) :- 
   write('(]= '), write(C1), write(' '), write(P), write(') ^ '),
   write('(=[ '), write(C2), write(' '), write(P), write(')'), !.
wax(subclass_of(X,Y)) :-  wax(X), write(' [[== '), wax(Y), !.
wax(disjoint_classes([X,Y])) :- wax(X), write(' ^ '), wax(Y), write(' = BOTTOM'), !.
wax(equivalent_classes([X,Y])) :- wax(X), write(' == '), wax(Y), !.
wax(object_property(P,List)) :- wprop(P,List), !.
wax(X) :- nl, write('***ERROR - CANT WRITE FORMULA '), write(X), nl, fail.

winter([X]) :- !, wax(X).
winter([X|Rest]) :- wax(X), write(' ^ '), winter(Rest).

wunion([X]) :- !, wax(X).
wunion([X|Rest]) :- wax(X), write(' U '), wunion(Rest).

wprop(_,[]) :- !.
wprop(P,[X|L]) :- write(property(P)), write(' - '), wp(X), nl, wprop(P,L).

   wp(domain(X)) :- !, write('domain('), wax(X), write(')').
   wp(range(X)) :- !, write('range('), wax(X), write(')').
   wp(X) :- write(X).

% formula to atom; make a temporary class unless it is already a named class

tostring(F,F) :-
   atom(F), !.
tostring(F,S) :-
   tos(F,Name),
   name(S,[36,36|Name]).

tos(X,S) :- atom(X), !, name(X,S).
tos(intersection_of(List),S) :- !, tosinter(List,S).  
tos(complement_of(X),[126|Rest]) :- !, name(X,Rest).
tos(restriction(P,[some_values_from(C)]),[40,69,32|Rest]) :- !,
   name(P,N1), tos(C,S), append(S,")",S1),
   append(N1,[46,32|S1],Rest).
tos(restriction(P,[all_values_from(C)]),[40,65,32|Rest]) :- !,
   name(P,N1), tos(C,S), append(S,")",S1),
   append(N1,[46,32|S1],Rest).
tos(restriction(P,[cardinality(C)]),[40,61,32|L]) :- !,
   name(C,Na1), name(P,Na2), append(Na2,")",L1),
   append(Na1,[32|L1],L).
tos(restriction(P,[min_cardinality(C)]),[40,93,61,32|L]) :- !,
   name(C,Na1), name(P,Na2), append(Na2,")",L1),
   append(Na1,[32|L1],L).
tos(restriction(P,[max_cardinality(C)]),[40,61,91,32|L]) :- !,
   name(C,Na1), name(P,Na2), append(Na2,")",L1),
   append(Na1,[32|L1],L).
tos(restriction(P,[min_cardinality(C1),max_cardinality(C2)]),[40,93,61,32|L]) :- !,
   name(C2,Na2), name(P,Na), 
   append(Na,")",L2),
   append(Na2,[32|L2],L3),
   name(C1,Na1),
   append(Na,[41,32,94,32,40,61,91,32|L3],L4),
   append(Na1,[32|L4],L).

tos(X,_) :- nl, write('***ERROR - CANT STRINGIFY FORMULA '), write(X), nl, fail.

tosinter([X],S) :- !, tos(X,S).
tosinter([X|Rest],S) :- tos(X,S1), tosinter(Rest,S2), append(S1,[32,94,32|S2],S).



