% Pre-processing OWL axioms
% Assumes axioms in axiom/1 format
%
% Main predicates provided:
%
% number_axioms - for each axiom(Ax), asserts a fact
%    axiom(N,Ax,M) where N is a unique number for the axiom
%    and M is the list of role and concept names in the axiom
% relevant(+C,-Axioms) - computes for a class or role name C
%    the list of numbers of the axioms relevant for explaining it 

% number axioms and record concepts and roles mentioned in them

number_axioms :- 
   retractall(axiom(_,_,_)),
   findall(A,axiom(A),List),
   na(1,List).

all_things(L) :- 
   setof(X,N^Ax^M^(axiom(N,Ax,M),member(X,M)),L).

:- dynamic axiom/3.

na(_,[]).
na(N,[A|Axs]) :-
   findall(C,mowl_axiom(A,C),L),
   assert(axiom(N,A,L)),
   N1 is N+1,
   na(N1,Axs).

% mowl_axiom(A,C) - C is mentioned in axiom A
%
% The definitions here follow owl_axiom (predicates have an initial
% "m" and an extra argument), though some simplications are possible

mowl_axiom(A,C) :- mowl_axiom(A,C,_).

mowl_class(A,C) :- mowl_axiom(A,C,class).

mowl_axiom(A,C,Ty) :- mowl_axiom(A,C,Ty,[]).

% first some non-OWL things we allow
mowl_axiom(not_subclass_of(D1,_),ID,T,O) :- mdescription(D1,ID,T,O).
mowl_axiom(not_subclass_of(_,D2),ID,T,O) :- mdescription(D2,ID,T,O).
% now proper OWL
mowl_axiom(class(ID,_,_),ID,class,_).
mowl_axiom(class(_,_,DList),ID,T,O) :-
   mdescription_list(DList,ID,T,O).
mowl_axiom(enumerated_class(ID,_,_),ID,class,_).
mowl_axiom(enumerated_class(_,IList),ID,T,O) :-
   mindividual_idlist(IList,ID,T,O).
mowl_axiom(disjoint_classes([D1,D2|DList]),ID,T,O) :- 
   mdescription_list([D1,D2|DList],ID,T,O).
mowl_axiom(equivalent_classes([D1,D2|DList]),ID,T,O) :- 
   mdescription_list([D1,D2|DList],ID,T,O).
   % OWL says can have just one!
mowl_axiom(subclass_of(D1,_),ID,T,O) :- mdescription(D1,ID,T,O).
mowl_axiom(subclass_of(_,D2),ID,T,O) :- mdescription(D2,ID,T,O).
mowl_axiom(datatype(ID),ID,datatype,_).
mowl_axiom(datatype_property(DPID,_),DPID,role,_).
mowl_axiom(datatype_property(P,DPL),ID,T,O) :-
   mdata_prop_list(DPL,P,ID,T,O).
mowl_axiom(object_property(DPID,_),DPID,role,_).
mowl_axiom(object_property(P,DPL),ID,T,O) :-
   mindiv_prop_list(DPL,P,ID,T,O).
mowl_axiom(equivalent_properties([D1,D2|DPIDs]),ID,T,O) :-
   mdata_prop_id_list([D1,D2|DPIDs],ID,T,O).   
mowl_axiom(subproperty_of(DPID1,_),DPID1,role,_).
mowl_axiom(subproperty_of(_,DPID2),DPID2,role,_).
mowl_axiom(equivalent_properties([D1,D2|DPIDs]),ID,T,O) :-
   mindiv_prop_id_list([D1,D2|DPIDs],ID,T,O).   
mowl_axiom(subproperty_of(DPID1,_),DPID1,role,_).
mowl_axiom(subproperty_of(_,DPID2),DPID2,role,_).
mowl_axiom(instance_of(I,_),I,individual,_).
mowl_axiom(instance_of(_,C),C,class,_).
mowl_axiom(related(I,_,_),I,individual,_).
mowl_axiom(related(_,P,_),P,role,_).
mowl_axiom(related(_,_,I),I,individual,_).

mdescription(A,B,C) :- mdescription(A,B,C,[]).

mdescription(ID,ID,class,_) :- class_id(ID).
mdescription(R,ID,T,O) :- mrestriction(R,ID,T,O).
mdescription(union_of(DList),ID,T,O) :- mdescription_list(DList,ID,T,O).
   % OWL says you can have any number!
mdescription(intersection_of(DList),ID,T,O) :- mdescription_list(DList,ID,T,O).
   % OWL says you can have any number!
mdescription(complement_of(D),ID,T,O) :- mdescription(D,ID,T,O).
mdescription(one_of(IDList),ID,T,O) :- mindividual_idlist(IDList,ID,T,O).
   % OWL says you can have zero!

mrestriction(restriction(DPID,L),DPID,role,O) :-
   (member(ronly,O) -> real_restr(L); true).
mrestriction(restriction(P,[D|DRCList]),ID,T,O) :-
    mrclist([D|DRCList],P,ID,T,O).

   real_restr(L) :- member(some_values_from(_),L), !.
   real_restr(L) :- member(all_values_from(_),L), !.

mrestriction_component(all_values_from(R),P,ID,T,O) :- look_inside(P,O),
   (mdata_range(R,ID,T,O); mdescription(R,ID,T,O)).
mrestriction_component(some_values_from(R),P,ID,T,O) :- look_inside(P,O),
   (mdata_range(R,ID,T,O); mdescription(R,ID,T,O)).
mrestriction_component(value(R),P,R,individual,O) :- look_inside(P,O).

   look_inside(P,O) :- member(toplevel(Rs),O), !, member(P,Rs), !.
   look_inside(_,_).   

mdata_range(D,D,datatype,_) :- datatype_id(D).

mdata_prop_elt(super(DPID),_,DPID,role,_).
mdata_prop_elt(domain(D),_,ID,T,O) :- mdescription(D,ID,T,O).
mdata_prop_elt(range(R),_,ID,T,O) :- mdata_range(R,ID,T,O).

mindiv_prop_elt(super(DPID),_,DPID,role,_).
mindiv_prop_elt(inverse_of(DPID),_,DPID,role,_).
mindiv_prop_elt(domain(D),_,ID,T,O) :- mdescription(D,ID,T,O).
mindiv_prop_elt(range(D),P,ID,T,O) :- 
   (member(toplevel(Rs),O) -> member(P,Rs); true), !,
   mdescription(D,ID,T,O).

% Lists

      mdescription_list([D|_],ID,T,O) :- mdescription(D,ID,T,O).
      mdescription_list([_|Ds],ID,T,O) :- mdescription_list(Ds,ID,T,O).

   mrclist([DRC|_],P,ID,T,O) :- mrestriction_component(DRC,P,ID,T,O).
   mrclist([_|DRCs],P,ID,T,O) :- mrclist(DRCs,P,ID,T,O).

      mindividual_idlist([ID|_],ID,individual,_).
      mindividual_idlist([_|IDs],ID,T,O) :- mindividual_idlist(IDs,ID,T,O).

   mdata_prop_list([P|_],R,ID,T,O) :- mdata_prop_elt(P,R,ID,T,O).
   mdata_prop_list([_|Ps],R,ID,T,O) :- mdata_prop_list(Ps,R,ID,T,O).

   mindiv_prop_list([P|_],R,ID,T,O) :- mindiv_prop_elt(P,R,ID,T,O).
   mindiv_prop_list([_|Ps],R,ID,T,O) :- mindiv_prop_list(Ps,R,ID,T,O).

   mdata_prop_id_list([ID|_],ID,role,_).
   mdata_prop_id_list([_|Ps],ID,T,O) :- mdata_prop_id_list(Ps,ID,T,O).

   mindiv_prop_id_list([ID|_],ID,role,_).
   mindiv_prop_id_list([_|Ps],ID,T,O) :- mindiv_prop_id_list(Ps,ID,T,O).

% Find relevant classes and properties for subsumers of class X, excluding X and anything 
% inside Others

rel_classes(r(Cl,_,_),Cl).
rel_props(r(_,P,_),P).
spec_props(r(_,_,P),P).

   find_relevant(X,Others,r(Classes,Props,Spec)) :-
      relevant(X,Axs),   
      % write('****RELAX: '), write(Axs), nl,
      once((setof(C,A^Ax^Me^(member(A,Axs),axiom(A,Ax,Me),mowl_axiom(Ax,C,class),\+ C=X,\+ member(C,Others)),Classes)
        ; Classes=[])),  
      length(Classes,N1), 
      (att_tracing -> (write('for '), write(X), nl, write(N1), write(' relevant classes: '), write(Classes), nl); true),
      % rec(rel_concepts,N1),
      once((setof(C,A^Ax^Me^(member(A,Axs),axiom(A,Ax,Me),mowl_axiom(Ax,C,role)),Props); Props=[])),
      length(Props,N2),
      (att_tracing -> (write(N2), write(' relevant properties: '), write(Props), nl); true),
      % rec(rel_properties,N2),
      % Spec=Props, N3=N2,
      special_roles(X,Spec), length(Spec,N3),
      (att_tracing -> (write(N3), write(' special properties: '), write(Spec), nl); true).
      % rec(spec_properties,N3).

% Determine relevant axioms for some class

relevant(C,Axioms) :-
   findall(N,M^A^(axiom(N,A,M),gci(A)),Ax),
   (setof(C1,Nu^A1^M^(member(Nu,Ax),axiom(Nu,A1,M),member(C1,M)),L1); L1=[]), !,
   findall(NC,mdescription(C,NC,class),L2),
   append(L1,L2,Names),
   rel(Names,[],Ax,_,Axioms).

% rel(ToScan,RelCSoFar,RelASofar,RelC,RelA).

rel([],R,A,R,A) :- !.
rel([C|Cs],R,A,R2,A2) :-
   member(C,R), !,
   rel(Cs,R,A,R2,A2).
rel([C|Cs],R,A,R2,A2) :-
   findall(Nu,(axiom(Nu,Ax,M),defines(Ax,C),\+member(Ax,A)),L),
   append(L,A,A1),
   (setof(C1,Nu^Ax^M^(member(Nu,L),axiom(Nu,Ax,M),member(C1,M)),C1s); C1s=[]), !,
   append(C1s,Cs,C2s),
   rel(C2s,[C|R],A1,R2,A2).

% special roles for some class

special_roles(C,Roles) :-
   findall(N,M^A^(axiom(N,A,M),gci(A)),Ax),
   (setof(C1,Nu^A1^M^(member(Nu,Ax),axiom(Nu,A1,M),member(C1,M)),L1); L1=[]), !,
   findall(NC,mdescription(C,NC,class),L2),
   append(L1,L2,Names),
   (setof(P,inversey_prop(P),Rs); Rs=[]), !,
   sprel(Names,Rs,[],Ax,Ment,Axioms),
   (setof(R,rolement(Ment,Axioms,R),Roles); Roles=[]), !.
  
      rolement(Ment,Axioms,R) :- 
         member(NN,Axioms), axiom(NN,AA,_), mowl_axiom(AA,R,role), member(R,Ment).

   inversey_prop(P) :- axiom(object_property(P,[inverse_of(_)])).
   inversey_prop(P) :- axiom(object_property(_,[inverse_of(P)])).
   inversey_prop(P) :- axiom(subproperty_of(P,P1)), inversey_prop(P1).
   inversey_prop(P) :- axiom(object_property(P,[super(P1)])), inversey_prop(P1).

% rel(ToScan,RelCSoFar,RelASofar,RelC,RelA).

sprel([],_,R,A,R,A) :- !.
sprel([C|Cs],Rs,R,A,R2,A2) :-
   member(C,R), !,
   sprel(Cs,Rs,R,A,R2,A2).
sprel([C|Cs],Rs,R,A,R2,A2) :-
   findall(Nu,(axiom(Nu,Ax,_),defines(Ax,C),\+member(Ax,A)),L),
   append(L,A,A1),
   (setof(C1,ment_ronly(L,Rs,C1),C1s); C1s=[]), !,
   append(C1s,Cs,C2s),
   sprel(C2s,Rs,[C|R],A1,R2,A2).

   ment_ronly(L,Rs,C1) :-
      member(Nu,L), axiom(Nu,Ax,_), mowl_axiom(Ax,C1,_,[ronly,toplevel(Rs)]).




% defines(Ax,C) - Axiom defines a concept/role

defines(class(ID,_,_),ID) :- atomic(ID).
defines(enumerated_class(ID,_),ID) :- atomic(ID).

% For the following, more complex descriptions will mean they are GCIs
defines(disjoint_classes(DList),ID) :- member(ID,DList), atomic(ID). 
defines(equivalent_classes(DList),ID) :- member(ID,DList), atomic(ID).
defines(subclass_of(D1,_),D1) :- atomic(D1).

defines(datatype_property(DPID,_),DPID).
defines(object_property(_,[inverse_of(P)]),P).
defines(object_property(DPID,_),DPID).        % also need to return inverses (not supers)?
defines(equivalent_properties(DPIDs),ID) :- member(ID,DPIDs).
defines(subproperty_of(DPID1,_),DPID1).

% general concept inclusion axioms

gci(class(ID,_,_)) :- \+ atomic(ID).
gci(enumerated_class(ID,_)) :- \+ atomic(ID).
gci(disjoint_classes(DList)) :- member(ID,DList), \+ atomic(ID). 
gci(equivalent_classes(DList)) :- member(ID,DList), \+ atomic(ID).
gci(subclass_of(D1,_)) :- \+ atomic(D1).




   
