% Simple realisation of OWL axioms
%
% Realisations are represented by structures ("trees") of the form Cat:List,
% where Cat is a syntactic category and List is a list of such
% structures and simple atoms (which are English words)
%
% Cat is usually an atom, but for sentences a category s(X) is used,
% where X is the OWL expression which gives rise to the subject of
% the sentence.

:- op(600,xfy,':').

:- [real].   % provides real/2

% output the best realisation. also record it

best_realise(Ax) :-
   best_realisation(Ax,Tree,_), 
   realise(Tree,List),
   record_realisation(Ax,List),
   writewords(List), nl, !.  

:- dynamic rec_real/2.

:- retractall(rec_real(_,_)).

   record_realisation(X,Y) :- rec_real(X,Y), !.
   record_realisation(X,Y) :- assertz(rec_real(X,Y)).

% best_realisation(+Formula,-Tree,-Eval)
%
% Given a formula, find the realisation with the least complexity (Tree),
% and its complexity (Eval)
% Fails if there is no realisation.

% special cases for things not in OWL or which span multiple sentences

best_realisation(although_not(Ax1,Ax2),T,Eval) :- 
   best_realisation(Ax1,Tree1,_), best_realisation(Ax2,s(S):[NP,VP],_),
   insert_adv([necessarily],VP,VP1),
   negate(VP1,VP2),
   T = s(asd):[conjn:[although],Tree1,conjn:[','],s(S):[NP,VP2]],
   complexity(T,0,Eval).
best_realisation(although_never(Ax1,Ax2),T,Eval) :- 
   best_realisation(Ax1,Tree1,_), best_realisation(Ax2,s(S):[NP,VP],_),
   insert_adv([never],VP,VP1),
   T = s(asd):[conjn:[although],Tree1,conjn:[','],s(S):[NP,VP1]],
   complexity(T,0,Eval).
best_realisation(not_subclass_of(X,Y),T,Eval) :- !,
   best_realisation(subclass_of(X,Y),s(X):[NP,VP],_),
   negate(VP,VP1),
   T = s(X):[NP,VP1],
   complexity(T,0,Eval).

best_realisation(Ax,T,Eval) :-
   findall(Complex-T1,(findn(5,real(Ax,T1)),complexity(T1,0,Complex)),List),
   \+ List=[], !,
   keysort(List,[Eval-T|_]),
   realise(T,L),
   record_realisation(Ax,L).

% Uncomment the following if you want to be told about formulae
% with no realisations
best_realisation(Ax,_,_) :-
   write('**** NO REALISATION FOR: '), writeq(Ax), nl, fail.

% Show realisation options

options(Ax) :-
   findall(Complex-T1,(real(Ax,T1),complexity(T1,0,Complex)),List),
   \+ List=[], !,
   keysort(List,Sorted),
   doall((member(C-T,Sorted),write('('),write(C),write(') '),realise(T,L),writewords(L),nl)).

% Operations on trees

% Take a tree and print out the actual sequence of words. Postprocessing
% takes place at this stage

realise(Str) :- realise(Str,List), writewords(List), nl.

writewords([]) :- !.
writewords([','|Ws]) :- !, write(','), writewords(Ws).
writewords([W|Ws]) :- write(' '), write(W), writewords(Ws).

realise(Tree,[First1|List]) :-
   realise(Tree,[vform=pres,num=sing],[First|List],[]),
   capitalise(First,First1). 

% realise(Tree,CxtIn,Words,EndOfWords)

% cross reference number
realise(C:['X'(Num)|L],Cx,L1,L2) :- !,
   member(num=Num,Cx), !,
   realise(C:L,Cx,L1,L2).

realise(s:[NP,VP],Cx,L1,L3) :- !,
   realise_np(NP,Cx,Num,L1,L2),
   realise(VP,[vform=pres,num=Num|Cx],L2,L3).
realise(s(_):[NP,VP],Cx,L1,L3) :- !,
   realise_np(NP,Cx,Num,L1,L2),
   realise(VP,[vform=pres,num=Num|Cx],L2,L3).

realise(np:L,Cx,L1,L2) :- !,
   realise_np(np:L,Cx,_,L1,L2).

% verb inflection
realise(vp:[aux:[can],VP],Cx,[can|L1],L2) :- !,
   realise(VP,[vform=base|Cx],L1,L2).
realise(vp:[aux:[to],VP],Cx,[to|L1],L2) :- !,
   realise(VP,[vform=base|Cx],L1,L2).
realise(vp:[aux:[do],VP],Cx,[DO|L1],L2) :- !,
   member(num=N,Cx), !,
   (N=sing -> DO=does; DO=do),
   realise(VP,[vform=base|Cx],L1,L2).
realise(verb:[V|L],Cx,L1,L2) :- !,
   member(vform=VF,Cx), member(num=N,Cx), !,
   ((VF=pres,N=sing) -> third_sing(V,VI); 
    ((VF=pres,N=plur) -> plur_verb(V,VI); VI=V)),
   append([VI|L],L2,L1).

% noun - pick up inflection added by determiner
realise(n:[N],Cx,[NI|L],L) :- !,
   member(num=X,Cx), !,
   (X=plur -> plural_noun(N,NI); NI=N).

% adverbs and "be" (applies to present tense only)
realise(vp:[adv:W,vp:[verb:[be|L],adv:[not]|Cs]],Cx,L1,L2) :- 
   once(member(vform=VF,Cx)), VF=pres, !,
   realise(vp:[verb:[be],adv:[not],adv:W,part:L|Cs],Cx,L1,L2).
realise(vp:[adv:W,vp:[verb:[be|L]|Cs]],Cx,L1,L2) :- 
   once(member(vform=VF,Cx)), VF=pres, !,
   realise(vp:[verb:[be],adv:W,part:L|Cs],Cx,L1,L2).

realise(_:List,Cx,X,Y) :- !, realise(List,Cx,X,Y).
realise([],_,X,X) :- !.
realise([X|Y],Cx,L1,L3) :- !, realise(X,Cx,L1,L2), realise(Y,Cx,L2,L3).
realise(X,_,[X|L],L).

% Realise NP - extra output specifies number

% indefinite article
realise_np(np:[det:[indef],nbar:[n:[thing]|Ms]],Cx,sing,[something|L1],L2) :- !,
   realise(Ms,[num=sing|Cx],L1,L2).  
realise_np(np:[det:[indef],nbar:[adj:[],nbar:[n:[thing]|Ms]]],Cx,sing,[something|L1],L2) :- !,
   realise(Ms,[num=sing|Cx],L1,L2).    
realise_np(np:[det:[indef],R],Cx,sing,[A|L1],L2) :- !,
   first_word(R,W),
   a_an(W,A),
   realise(R,[num=sing|Cx],L1,L2).
realise_np(np:[det:[any],nbar:[n:[thing]|Ms]],Cx,sing,[anything|L1],L2) :- !,
   realise(Ms,[num=sing|Cx],L1,L2).  

% pronoun 
realise_np(np:[det:[],nbar:[n:[pron(Num)]]],_,Num,[Pron|L],L) :- !,
   pronoun(Num,Pron).

% other determiners set number
realise_np(np:[det:W,NB],Cx,Num,L1,L3) :- !,
   realise(W,Cx,L1,L2),
   determiner_num(W,Num),
   realise(NB,[num=Num|Cx],L2,L3).

% other NP cases

realise_np(np:[adv:A,NP],Cx,Num,L1,L3) :- !,
   realise(adv:A,Cx,L1,L2),
   realise_np(NP,Cx,Num,L2,L3).
realise_np(np:[conjn:C1,np:N1,conjn:C2,np:N2],Cx,plur,L0,L4) :- !,
   realise(conjn:C1,Cx,L0,L1),
   realise_np(np:N1,Cx,_,L1,L2), 
   realise(conjn:C2,Cx,L2,L3),
   realise_np(np:N2,Cx,_,L3,L4).
realise_np(np:[np:NP1,conjn:C,np:NP2],Cx,plur,L1,L4) :- !,
   realise_np(np:NP1,Cx,_,L1,L2), 
   realise(conjn:C,Cx,L2,L3),
   realise_np(np:NP2,Cx,_,L3,L4).
realise_np(np:[trace:[]],Cx,Num,L,L) :-   % trace takes its number from enclosing context
   member(num=Num,Cx), !.

writelist([X]) :- !, write(X), write('.'), nl.
writelist([X|L]) :- write(X), write(' '), writelist(L).

    capitalise(W1,W2) :- name(W1,[C|N1]), C=<"z", C>="a", !,
       C1 is C-"a"+"A", name(W2,[C1|N1]).  
    capitalise(W,W).

   % first word in a tree
   first_word(_:L,W) :- !, member(X,L), first_word(X,W), !.
   first_word(X,X).

    % choose a determiner, given the following noun
    a_an(C,an) :- name(C,[C1|_]), vowel(C1), !.
    a_an(_,a).

    vowel(C) :- (C is "a"; C is "e"; C is "i"; C is "o"; C is "u";
                C is "A"; C is "E"; C is "I"; C is "O"; C is "U"), !.

% number of a determiner

determiner_num([at,most,N],plur) :- integer(N), N>1, !.
determiner_num([at,least,N],plur) :- integer(N), N>1, !.
determiner_num([exactly,N],plur) :- integer(N), N>1, !.
determiner_num(Ws,N) :- append(_,[and|Rest],Ws), !,
   determiner_num(Rest,N).
determiner_num([no],plur) :- !.
determiner_num([any],sing) :- !.    % any only used with singulars
determiner_num(_,sing).

% third person singular of a verb

third_sing(be,is) :- !.
third_sing(have,has) :- !.
third_sing(do,does) :- !.
third_sing(V,VI) :- name(V,VN), append(VN,"s",VIN), name(VI,VIN), !.

% plural form of a verb

plur_verb(be,are) :- !.
plur_verb(V,V).

% plural form of a noun

plural_noun(child,children) :- !.
plural_noun(person,people) :- !.
plural_noun(N,NI) :- name(N,Na), append(Na,"s",Na1), name(NI,Na1).

% pronoun

pronoun(sing,it) :- !.
pronoun(plur,they).

% complexity(+Tree,+Sofar,-Complex)
%
% Compute the complexity of a tree (ie. number of words) Complex,
% added to the value Sofar which has already been worked out.

complexity(_:List,N,N1) :- !, complexity(List,N,N1).
complexity([],N,N) :- !.
complexity([X|Y],N1,N3) :- !, complexity(X,N1,N2), complexity(Y,N2,N3).
complexity(_,N,N1) :- N1 is N+1.


% The grammar

rule(s,[np,vp]).
rule(s,[s,conj,s]).
rule(vp,[vp,conj,vp]).
rule(vp,[conj,vp,conj,vp]).
rule(vp,[adv,vp]).
rule(vp,[aux,vp]).
rule(vp,[verb,*(np;pp;adj;adv)]).
rule(pp,[prep,np]).
rule(np,[det,nbar]).
rule(nbar,[adj,nbar]).
rule(nbar,[n,*(pp;relc;adj)]).
rule(relc,[relpron,s]).
rule(np,[adv,np]).
rule(np,[np,conj,np]).
rule(np,[conj,np,conj,np]).
rule(np,[trace]).

check_syn(C:List) :-
   rule(C,_), !,
   rule(C,L), check_rhs(L,List).
check_syn(_).

check_rhs([],[]) :- !.
check_rhs([*(_)|L],List) :- 
   check_rhs(L,List).
check_rhs([*(Disj)|L],[C:X|List]) :- !,   
   dmember(C,Disj), check_syn(C:X), check_rhs([*(Disj)|L],List).
check_rhs([C|L],[C:Y|L1]) :- 
   check_syn(C:Y), check_syn(L,L1).

       dmember(X,(A;_)) :- dmember(X,A).
       dmember(X,(_;B)) :- !, dmember(X,B).
       dmember(X,X).

cr(Ax) :- findall(Z,real(Ax,Z),L), forall(member(Z,L),(check_syn(Z)->(write(ok),nl);(write(Z), nl))).

% Display a tree

dtree(T) :- dtree(0,T).

dtree(N,C:L) :- !, tab(N), write(C),
   (member(_:_,L) -> (nl, N1 is N+2, dtree(N1,L)); (write(' '), write(L), nl)).
dtree(_,[]) :- !.
dtree(N,[X|L]) :- dtree(N,X), dtree(N,L).

% Grammar sensitive operations

% Get main verb of a VP
main_verb(vp:[verb:L|_],L1) :- !, L=L1.
main_verb(vp:[aux:_,VP],V) :- !, main_verb(VP,V).
main_verb(vp:[adv:_,VP],V) :- !, main_verb(VP,V).

% Inserting an adverb in a VP. Adverbs come after auxs

insert_adv(Ws,vp:[aux:A,X],vp:[aux:A,X1]) :- !, insert_adv(Ws,X,X1).  % after any auxs
insert_adv(Ws,VP,vp:[adv:Ws,VP]).

insert_aux(Ws,VP,vp:[aux:Ws,VP]).

negate(vp:['X'(X)|R],vp:['X'(X)|R1]) :- !, negate(vp:R,vp:R1).
negate(vp:[aux:A,VP],vp:[aux:A,vp:[aux:[not],VP]]) :- !.
negate(vp:[adv:A,VP],vp:[adv:A,VP1]) :- main_verb(VP,[be|_]), !, negate(VP,VP1).
negate(vp:[adv:A,VP],vp:[aux:[do],[adv:[not],vp:[adv:A,VP]]]) :- !.
negate(vp:[verb:[be|X]|Cs],vp:[verb:[be],adv:[not],part:X|Cs]) :- !.
negate(vp:[verb:V|Cs],vp:[aux:[do],vp:[adv:[not],vp:[verb:V|Cs]]]).

